/**
 * @file
 * @brief Decoupling thresholds
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.3
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015 Andreas Maier
 *
 * This file is part of the QCD library.
 *
 * The QCD library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QCD library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QCD library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#pragma once
/* define decoupling thresholds */

#include <set>
#include <initializer_list>
#include <limits>
#include <cmath>

#include "scheme.hpp"
#include "masses.hpp"

namespace QCD{

  //!Threshold for the decoupling of a quark flavour
  struct threshold{
    //! Decoupling scale
    double mu;
    //! Mass of the quark to be decoupled
    double m;
    //! Scheme of the quark mass
    scheme dec_scheme;

    //! Threshold for a massless quark (no decoupling)
    inline constexpr threshold()
      : mu(0.), m(0.), dec_scheme(scheme::pole) {}

    //! Construct a decoupling threshold
    /**
     * @param scale   Decoupling scale
     * @param mass    Mass of the quark to be decoupled
     * @param dec     Scheme of the quark mass
     */
    inline constexpr threshold(
	double scale,
	double mass,
	scheme dec = scheme::pole
    ): mu(scale), m(mass), dec_scheme(dec) {}
  };

  //! Default charm threshold
  static constexpr threshold charm_threshold{2*mc_OS, mc_OS};
  //! Default bottom threshold
  static constexpr threshold bottom_threshold{mb_OS, mb_OS};
  //! Default top threshold
  static constexpr threshold top_threshold{mt_OS, mt_OS};

  inline bool operator<(threshold const & a,threshold const & b){
    return a.mu < b.mu;
  }

}
