/**
 * @file
 * @brief Gamma function
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.4
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#pragma once

#include "complex.hpp"
#include "zeta.hpp"
#include <array>
#include <cmath>

namespace Gamma{
  using namespace complex;
  using zeta::pi;

  /*
   * coefficients from taken from gsl
   * see below for the algorithm
   */
  static constexpr int g = 7;
  static constexpr auto p = std::array<double,9>{
    0.99999999999980993,
    676.5203681218851,
    -1259.1392167224028,
    771.32342877765313,
    -176.61502916214059,
    12.507343278686905,
    -0.13857109526572012,
    9.9843695780195716e-6,
    1.5056327351493116e-7
  };

  //! Gamma function for complex argument
  /**
   *  Calculates the Gamma function using the Lanczos approximation
   */
  template<typename T>
  inline std::complex<T> Gamma(std::complex<T> z){
    if(real(z) < 1./2.) return pi/(sin(pi*z)*Gamma(1-z));
    z -= 1;
    std::complex<T> x = p[0];
    for(size_t i = 1; i < p.size(); ++i){
      x += p[i]/(z+i);
    }
    std::complex<T> const t = z + g + 1./2.;
    return sqrt(2*pi)*pow(t,z + 1./2.)*exp(-t)*x;
  }

  //! Gamma functions for real argument
  /**
   *  Falls back to the standard library implementation
   */
  inline float Gamma(float z){
    return std::tgamma(z);
  }

  inline double Gamma(double z){
    return std::tgamma(z);
  }

  inline long double Gamma(long double z){
    return std::tgamma(z);
  }

  inline double Gamma(int z){
    return std::tgamma(z);
  }
}
