/**
 * @file
 * @brief Hypergeometric functions
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.4
 *
 * @section LICENSE
 *
 * Copyright 2014 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#pragma once

#include <complex>
#include <algorithm>
#include <string>
#include <cassert>

#include "array.hpp"
#include "complex.hpp"
#include "Mellin_Li.hpp"
#include "psi.hpp"
#include "not_implemented.hpp"
#include "integrate.hpp"

#include "harmonic_sums.hpp"
#include "Gamma.hpp"
#include "zeta.hpp"
#include "dilog.hpp"

namespace hypergeometric{
  typedef std::complex<double> complex;
  using namespace complex;
  using namespace zeta;
  using psi::psi;
  using harmonic_sums::harmonic_sum;


  namespace detail{
    template<int n, typename T>
    struct equals{
      bool operator()(T t){
	return t == T{n};
      }
    };
  }

  /** \internal
   * Calculate hypergeometric function using Mellin-type integration
   */
  template<size_t P>
  double PFQ_int(
      std::array<double, P> const & p,
      std::array<double, P-1> const & q,
      double z
  ){
    using Gamma::Gamma;
    /* determine contour shift */
    const double ep = *std::min_element(p.begin(), p.end())/2;
    if(ep <= 0){
      throw QQbar_threshold::not_implemented(
	  "numerical integration of PFQ with non-positive p parameters"
      );
    }
    auto integrand = [&p, &q, z, ep](double t){
      const double s = tan(pi/2*t);
      complex result = pow(-z,complex{-ep,s})/pow(cos(pi/2*t),2)
      *Gamma(complex{ep, -s})*Gamma(complex{p[P-1]-ep, s})
      ;
      for(size_t i = 0; i < P-1; ++i){
	result *= Gamma(complex{p[i]-ep, s})/Gamma(complex{q[i]-ep, s});
      }
      return real(result);
    };
    auto i = integrate::make_integrator(integrand);
    double result = i.integrate(-1, 1)/(4*Gamma(p[P-1]));
    for(size_t i = 0; i < P-1; ++i){
      result *= Gamma(q[i])/Gamma(p[i]);
    }
    return result;
  }

  template<int P>
  double PFQ(
      std::array<double, P> const & p,
      std::array<double, P-1> const & q,
      double z
  ){
    return PFQ_int<P>(p, q, z);
  }

  /** \internal
   * String representation of hypergeometric function
   * Useful for error messages
   */
  template<typename iterator, typename arg_type>
  std::string PFQ_to_str(
      iterator p_begin, iterator p_end,
      iterator q_begin, iterator q_end,
      arg_type z
  ){
    std::string result = "PFQ({";
    for(auto it = p_begin; it != p_end; ++it){
      if(it != p_begin) result += ", ";
      result += to_string<double>(*it);
    }
    result += "}, {";
    for(auto it = q_begin; it != q_end; ++it){
      if(it != q_begin) result += ", ";
      result += to_string<double>(*it);
    }
    result += "}, " + to_string<double>(z) + ")";
    return result;
  }

  template<>
  inline double PFQ<2>(
      std::array<double, 2> const & p,
      std::array<double, 1> const & q,
      double z
  ){
    if(
  	std::all_of(p.cbegin(),p.cend(), detail::equals<1, complex>())
  	&& (q[0] == 7./2.)
	&& (z <= 0)
    ){
      z = -z;
      return - (20./3.*z + 5)/pow(z,2)
	+ 5*asinh(sqrt(z))*pow(1/z + 1, 3./2.)/z;
    }
    return PFQ_int<2>(p, q, z);
  }

  template<>
  inline double PFQ<3>(
      std::array<double, 3> const & p,
      std::array<double, 2> const & q,
      double z
  ){
    using dilog::Li2;
    static const auto p_ = {1., 1., 3./2.};
    static const auto q_ = {5./2., 7./2.};
    if(
  	std::equal(p.cbegin(), p.cend(), p_.begin())
  	&& std::equal(q.cbegin(), q.cend(), q_.begin())
	&& (z <= 0)
    ){
      assert(z <= 0);
      z = -z;
      return
  	+ (15./2. - 35*z + 135./4.*zeta_2*sqrt(z))/pow(z,2)
  	+ pow(z, -3./2.)*(
  	    + 45./4.*Li2(pow(sqrt(z) + sqrt(1 + z),-2))
  	    - 45*Li2(1/(sqrt(z) + sqrt(1 + z)))
  	    + asinh(sqrt(z))*(
  	    	+ 15*sqrt(1 + z)*(1 - 1/(2*z))
  	    	+ 45./2.*log(sqrt(1+z) - 1)
		- 45./4.*log(z)
  	    )
  	);
    }
    return PFQ_int<3>(p, q, z);
  }

  namespace detail{
    /*   coefficients in series expansion of      */
    /*   4F3({1, 1, 3/2, 3/2}, {5/2, 5/2, 7/2})   */
    constexpr double gen_an_4F3(int n){
      return (n==0)?1.:
	n*(n+1./2.)*(n+1./2.)/((n+3./2.)*(n+3./2.)*(n+5./2.))*gen_an_4F3(n-1);
    }

    auto constexpr an_4F3 = array::gen_array<500>(gen_an_4F3);

    /* evaluate 4F3({1, 1, 3/2, 3/2}, {5/2, 5/2, 7/2}) */
    inline
    double inner_4F3(double z){ /* through series expansion */
      const size_t n_max = std::nearbyint(10./(1-std::abs(z)));
      assert(n_max <= an_4F3.size());
      double result = 0;
      for(size_t n = n_max; n > 0; --n){
	result = z*(result + an_4F3[n]);
      }
      return result + an_4F3[0];
    }
  }

  template<>
  inline double PFQ<4>(
      std::array<double, 4> const & p,
      std::array<double, 3> const & q,
      double z
  ){
    static const auto p_ = {1., 1., 3./2., 3./2.};
    static const auto q_ = {5./2., 5./2., 7./2.};
    static constexpr double inner = 0.98; //define inner region for evaluation
    if(
  	std::equal(p.cbegin(), p.cend(), p_.begin())
  	&& std::equal(q.cbegin(), q.cend(), q_.begin())
	&& (std::abs(z) < inner)
    ){
      return detail::inner_4F3(z);
    }
    return PFQ_int<4>(p, q, z);
  }

  //! Analytic continuation of the generalized hypergeometric function
  /**
   * @returns  The value of PFQ({a_1,...,a_p2},{b_1,...,b_(p-1)},z)
   */
  template<int P>
  complex PFQ(
      std::array<complex, P> const & p,
      std::array<complex, P-1> const & q,
      int z
  ){
    if(
	(z == 1)
	&& std::all_of(p.cbegin(),p.cend(), detail::equals<1, complex>())
	&& std::all_of(q.cbegin(),q.cend()-1, detail::equals<2, complex>())
    ){
      complex const n = q.back();
      return (n-1) * detail::Mellin_Li<P-2>{}(n-2);
    }
    static const auto p_ = {1., 1., 4., 4.};
    static const auto q_ = {5., 5.};
    if(
	(P == 4)
	&& (z == 1)
	&& std::equal(p.cbegin(), p.cend(), p_.begin())
	&& std::equal(q.cbegin(), q.cend()-1, q_.begin())
    ){
      complex const lam = 3 - q.back();
      return 4*(lam - 2)*(lam - 1)*lam*(
	  2./3.*PFQ<4>({1,1,1,1},{2,2,-lam},1)
	  + (
	      (lam*(lam*(3*lam*(8 - 17*lam) - 20) + 11) - 18)
	      /(pow((lam - 1)*lam,2))
	      - 33*(1 + lam)*psi<1>(2-lam)
	  )/27.
      );
    }
    throw QQbar_threshold::not_implemented(
	PFQ_to_str(begin(p), end(p), begin(q), end(q), z)
    );
  }

  //! Derivative of hypergeometric functions with respect to last index
  /**
   *   der_PFQ({1,...},{2,...,z},1) = d/dz PFQ({1,...}, {2,..., z}, 1)
   */
  template<int P>
  complex der_PFQ(
      std::array<complex, P> const & p,
      std::array<complex, P-1> const & q,
      int z
  );

  //! Second derivative of hypergeometric functions with respect to last index
  /**
   *   der_PFQ({1,...},{2,...,z},1) = (d/dz)^2 PFQ({1,...}, {2,..., z}, 1)
   */
  template<int P>
  complex der2_PFQ(
      std::array<complex, P> const & p,
      std::array<complex, P-1> const & q,
      int z
  );

  namespace detail{
    template<int P> complex der_PFQ(complex z);
    template<> complex der_PFQ<4>(complex z);
    template<> complex der_PFQ<5>(complex z);

    template<int P> complex der2_PFQ(complex z);
    template<> complex der2_PFQ<4>(complex z);
  }

  template<int P>
  complex der_PFQ(
      std::array<complex, P> const & p,
      std::array<complex, P-1> const & q,
      int z
  ){
    if(
	(z == 1)
	&& std::all_of(p.cbegin(),p.cend(), detail::equals<1, complex>())
	&& std::all_of(q.cbegin(),q.cend()-1, detail::equals<2, complex>())
    ){
      return detail::der_PFQ<P>(q.back());
    }
    std::string error = "derivative of hypergeometric PFQ({";
    for(size_t i = 0; i < P-1; ++i){
      error += to_string(p[i]) + ", ";
    }
    error += to_string(p[P-1]) + "}, {";
    for(size_t i = 0; i < P-2; ++i){
      error += to_string(q[i]) + ", ";
    }
    error +=
      to_string(q[P-2]) + "}, "
      + std::to_string(z) + ")";
    throw QQbar_threshold::not_implemented(error);
  }

  template
  complex der_PFQ<4>(
      std::array<complex, 4> const & p,
      std::array<complex, 3> const & q,
      int z
  );

  template
  complex der_PFQ<5>(
      std::array<complex, 5> const & p,
      std::array<complex, 4> const & q,
      int z
  );

  template<int P>
  complex der2_PFQ(
      std::array<complex, P> const & p,
      std::array<complex, P-1> const & q,
      int z
  ){
    if(
	(z == 1)
	&& std::all_of(p.cbegin(),p.cend(), detail::equals<1, complex>())
	&& std::all_of(q.cbegin(),q.cend()-1, detail::equals<2, complex>())
    ){
      return detail::der2_PFQ<P>(q.back());
    }
    std::string error = "second derivative of hypergeometric PFQ({";
    for(size_t i = 0; i < P-1; ++i){
      error += to_string(p[i]) + ", ";
    }
    error += to_string(p[P-1]) + "}, {";
    for(size_t i = 0; i < P-2; ++i){
      error += to_string(q[i]) + ", ";
    }
    error +=
      to_string(q[P-2]) + "}, "
      + std::to_string(z) + ")";
    throw QQbar_threshold::not_implemented(error);
  }

  template
  complex der2_PFQ<4>(
      std::array<complex, 4> const & p,
      std::array<complex, 3> const & q,
      int z
  );

  namespace detail{
    template<>
    inline complex der_PFQ<4>(complex z){
      complex const lam = 1 - z;
      return
	+ zeta_3 + 1./2.*lam*pow(zeta_2,2)
	- harmonic_sum<2>(-1-lam)*lam*zeta_2
	+ 1./2.*pow(harmonic_sum<2>(-1-lam),2)*lam
	- harmonic_sum<2,1>(-1-lam)
	+ harmonic_sum<3>(-1-lam)
	- 2*harmonic_sum<3,1>(-1-lam)*lam
	+ 5./2.*harmonic_sum<4>(-1-lam)*lam
	+ harmonic_sum<1>(-1-lam) * (
	    - zeta_2
	    - 2*lam*zeta_3
	    + harmonic_sum<2>(-1-lam)
	    + 2*harmonic_sum<3>(-1-lam)*lam
	);
    }

    template<>
    inline complex der_PFQ<5>(complex z){
      complex const lam = 1 - z;
      return
	+ zeta_4
	+ pow(harmonic_sum<1>(-1-lam),2)/12*(pow(pi,2)-6*harmonic_sum<2>(-1-lam))
	+ pow(pi,2)*harmonic_sum<2>(-1-lam)/12
	- harmonic_sum<4>(-1-lam)/2
	+ harmonic_sum<3,1>(-1-lam)
	- harmonic_sum<2,1,1>(-1-lam)
	- harmonic_sum<1>(-1-lam)*(
	    harmonic_sum<3>(-1-lam) - harmonic_sum<2,1>(-1-lam) + zeta_3
	)
	+ lam*(
	    + harmonic_sum<3>(-1-lam)/6*(pow(pi,2)-12*harmonic_sum<2>(-1-lam))
	    - 3*harmonic_sum<5>(-1-lam)
	    - harmonic_sum<1>(-1-lam)/72*(
		+ pow(pi,4) - 12*pow(pi,2)*harmonic_sum<2>(-1-lam)
		+ 36*pow(harmonic_sum<2>(-1-lam),2)
		+ 180*harmonic_sum<4>(-1-lam)
		- 144*harmonic_sum<3,1>(-1-lam)
	    )
	    + 2*harmonic_sum<3,2>(-1-lam)
	    + 2*harmonic_sum<4,1>(-1-lam)
	    + harmonic_sum<2,2,1>(-1-lam)
	    - 2*harmonic_sum<3,1,1>(-1-lam)
	    - zeta_2*zeta_3
	    + pow(harmonic_sum<1>(-1-lam),2)*(-harmonic_sum<3>(-1-lam)+zeta_3)
	    + 3*zeta_5
	);
    }

    template<>
    inline complex der2_PFQ<4>(complex z){
      complex const lam = 1 - z;
      return
	- pow(pi,4)/36
	- pow(harmonic_sum<2>(-1-lam),2)
	+ lam*pow(pi,2)*harmonic_sum<3>(-1-lam)/3
	- 5*harmonic_sum<4>(-1-lam)
	- 10*lam*harmonic_sum<5>(-1-lam)
	+ 4*harmonic_sum<3,1>(-1-lam)
	+ 2*lam*harmonic_sum<3,2>(-1-lam)
	+ 6*lam*harmonic_sum<4,1>(-1-lam)
	- lam*pow(pi,2)*zeta_3/3
	+ harmonic_sum<1>(-1-lam)*(
	    + lam*pow(pi,4)/15
	    - 4*harmonic_sum<3>(-1-lam)
	    - 6*lam*harmonic_sum<4>(-1-lam)
	    + 4*zeta_3
	)
	+ harmonic_sum<2>(-1-lam)*(
	    pow(pi,2)/3 - 4*lam*harmonic_sum<3>(-1-lam) + 4*lam*zeta_3
	)
	+ lam*zeta_5;
    }
  }

};
