/**
 * @file
 * @brief Grids with interpolation between points
 *
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.4
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */

#pragma once

#include <algorithm>
#include <cassert>
#include <vector>
#include <array>

namespace QQbar_threshold{
  namespace detail{

    //! Two-dimensional matrix
    /**
     * @tparam T Element type
     */
    template<typename T>
    class matrix{
    public:
      //! Construct empty matrix
      matrix(){}

      //! Construct matrix with fixed number of rows and columns
      /**
       * @param rows     number of rows
       * @param columns   number of columns
       */
      matrix(size_t rows, size_t columns);

      //! Construct matrix with given entries
      /**
       * @param  columns   number of columns
       * @params entries   matrix entries
       */
      matrix(
          size_t columns,
          std::initializer_list<T> entries
      );

      //! Access matrix element at given position
      T & operator()(size_t row, size_t col);

      //! Access matrix element at given position
      T const & operator()(size_t row, size_t col) const;

      //! Return number of rows
      size_t rows() const;

      //! Return number of columns
      size_t columns() const;

    private:
      size_t rows_;
      size_t columns_;
      std::vector<T> entries_;
    };

    //! Grid with interpolation
    template<typename T>
    class interpolation_grid{
    public:
      interpolation_grid() = default;

      //! Construct grid with given coordinates and entries
      /**
       * @param x_coord    x coordinates of grid points
       * @param y_coord    y coordinates of grid points
       * @param values     values associated with grid points
       */
      interpolation_grid(
          std::vector<double> x_coord,
          std::vector<double> y_coord,
          matrix<T> values
      );

      //! Minimum x coordinate
      double x_min() const;
      //! Maximum x coordinate
      double x_max() const;

      //! Minimum y coordinate
      double y_min() const;
      //! Maximum y coordinate
      double y_max() const;

      //! Check whether grid is empty
      bool empty() const;

      //! Interpolated grid value
      /**
       * @returns The interpolated value at the point (x,y)
       */
       T operator()(double x, double y) const;

      T interpolate_bilinear(double x, double y) const;

    private:
    //! Get grid point for continuous coordinate
    /**
     * @returns         The index of the next grid point not greater
     *                  than the given coordinate
     *
     * e.g. if the points of the grid have x coordinates
     * {..., -0.0655977, -5.00506e-16, 0.0655977, 0.131195, 0.196793, ...}
     * get_idx(x, 0) would return the index of -5.00506e-16
     *
     */
      size_t get_idx(std::vector<double> const & grid_coord, double coord) const;

      //! Interpolated grid value for given cell
      /**
       * @returns The interpolated value at the point (x,y),
       * where (x_idx, y_idx) is the next grid point with smaller values
       * of the coordinates.
       */
      T interpolate(
          double x_, double y_,
          size_t x_idx, size_t y_idx
      ) const;

      T interpolate_bilinear(
          double x_, double y_,
          size_t x_idx, size_t y_idx
      ) const;

      //initialise from a given grid f
      void init(matrix<T> f);

      //! Derivatives in y direction
      /**
       * @param f   Grid of function values
       * @returns   A grid with estimated values for df/dy
       */
      matrix<T> y_derivatives(matrix<T> f) const;

      //! Derivatives in x direction
      /**
       * @param f   Grid of function values
       * @returns   A grid with estimated values for df/dx
       */
      matrix<T> x_derivatives(matrix<T> f) const;

      //! initialise polynomial coefficients
      void init_c(matrix<T> f, matrix<T> fx, matrix<T> fy, matrix<T> fxy);
      //! \internal verifies that we indeed interpolate f
      bool interpolates(matrix<T> const & f) const;

      std::vector<double> x;
      std::vector<double> y;
      matrix<std::array< std::array<T, 4>, 4>> c;
    };

  }
}
