/**
 * @file
 * @brief Grids with interpolation between points (implementation)
 *
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.4
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */

#pragma once

#include <algorithm>
#include <cassert>
#include <vector>
#include <array>
#include <cmath>

#include "complex.hpp"

namespace QQbar_threshold{
  namespace detail{
    template<typename T>
    matrix<T>::matrix(size_t rows, size_t columns):
	rows_(rows),
	columns_(columns),
	entries_{std::vector<T>(rows * columns)}
      {}

    template<typename T>
    matrix<T>::matrix(
	size_t columns,
	std::initializer_list<T> entries
    ):
      rows_(entries.size()/columns),
      columns_(columns),
      entries_(entries)
    {}

    template<typename T>
    T & matrix<T>::operator()(size_t row, size_t col){
      return entries_[row*columns() + col];
    }

    template<typename T>
    T const & matrix<T>::operator()(size_t row, size_t col) const{
      return entries_[row*columns() + col];
    }

    template<typename T>
    size_t matrix<T>::rows() const{
      return rows_;
    }

    template<typename T>
    size_t matrix<T>::columns() const{
      return columns_;
    }

    template<typename T>
    interpolation_grid<T>::interpolation_grid(
	std::vector<double> x_coord,
	std::vector<double> y_coord,
	matrix<T> values
    ):
      x(std::move(x_coord)),
      y(std::move(y_coord)),
      c(values.rows()-1, values.columns()-1)
    {
      assert(std::is_sorted(begin(x), end(x)));
      assert(std::is_sorted(begin(y), end(y)));
      assert(values.rows() == x.size());
      assert(values.columns() == y.size());
      init(std::move(values));
    }

    template<typename T>
    inline void interpolation_grid<T>::init(matrix<T> f){
      matrix<T> fx = x_derivatives(f);
      matrix<T> fy = y_derivatives(f);
      matrix<T> fxy = y_derivatives(fx); //or x_derivatives(fy)
      init_c(f, fx, fy, fxy);
    }

    template<typename T>
    inline matrix<T> interpolation_grid<T>::y_derivatives(matrix<T> f) const{
      assert(y.size() == f.columns());
      matrix<T> fy(f.rows(), f.columns());
      const size_t N = f.columns()-1;
      assert(N > 0);
      for(size_t i = 0; i < f.rows(); ++i){
	fy(i, 0) = (f(i, 1) - f(i, 0))/(y[1] - y[0]);
	for(size_t j = 1; j < N; ++j){
	  fy(i, j) = (f(i, j+1) - f(i, j-1))/(y[j+1] - y[j-1]);
	}
	fy(i, N) = (f(i, N) - f(i, N-1))/(y[N] - y[N-1]);
      }
      return fy;
    }

    template<typename T>
    inline matrix<T> interpolation_grid<T>::x_derivatives(matrix<T> f) const{
      assert(x.size() == f.rows());
      matrix<T> fx(f.rows(), f.columns());
      const size_t N = f.rows()-1;
      assert(N > 0);
      for(size_t j = 0; j < f.columns(); ++j){
	fx(0, j) = (f(1, j) - f(0, j))/(x[1] - x[0]);
	for(size_t i = 1; i < N; ++i){
	  fx(i, j) = (f(i+1, j) - f(i-1, j))/(x[i+1] - x[i-1]);
	}
	fx(N, j) = (f(N, j) - f(N-1, j))/(x[N] - x[N-1]);
      }
      return fx;
    }

    template<typename T>
    inline void interpolation_grid<T>::init_c(
	matrix<T> f, matrix<T> fx, matrix<T> fy, matrix<T> fxy
    ){
      using namespace complex;
      std::array<std::array<T, 2>, 2> Fx, Fy, Fxy;
      for(size_t i = 0; i < c.rows(); ++i){
	for(size_t j = 0; j < c.columns(); ++j){
	  // auxiliary variables
	  for(size_t a = 0; a <= 1; ++a){
	    for(size_t b = 0; b <= 1; ++b){
	      Fx[a][b] = (x[i+1]-x[i])*fx(i+a,j+b);
	      Fy[a][b] = (y[j+1]-y[j])*fy(i+a,j+b);
	      Fxy[a][b] = (x[i+1]-x[i])*(y[j+1]-y[j])*fxy(i+a,j+b);
	    }
	  }
	  /*
	   * for the bicubic polynomial in the form given by interpolate()
	   * determine the coefficients c via the conditions
	   * p == f, px == fx, py == fy, pxy == fxy
	   * at the corners of the current grid cell
	   */
	  c(i, j)[0][0] = f(i,j);
	  c(i, j)[0][1] = f(i,1 + j);
	  c(i, j)[0][2] = f(i,j) - f(i,1 + j) + Fy[0][0];
	  c(i, j)[0][3] = -f(i,j) + f(i,1 + j) - Fy[0][1];
	  c(i, j)[1][0] = f(1 + i,j);
	  c(i, j)[1][1] = f(1 + i,1 + j);
	  c(i, j)[1][2] = f(1 + i,j) - f(1 + i,1 + j) + Fy[1][0];
	  c(i, j)[1][3] = -f(1 + i,j) + f(1 + i,1 + j) - Fy[1][1];
	  c(i, j)[2][0] = f(i,j) - f(1 + i,j) + Fx[0][0];
	  c(i, j)[2][1] = f(i,1 + j) - f(1 + i,1 + j) + Fx[0][1];
	  c(i, j)[2][2] = f(i,j) - f(i,1 + j) - f(1 + i,j) + f(1 + i,1 + j) + Fx[0][0] - Fx[0][1] + Fxy[0][0] + Fy[0][0] - Fy[1][0];
	  c(i, j)[2][3] = -f(i,j) + f(i,1 + j) + f(1 + i,j) - f(1 + i,1 + j) - Fx[0][0] + Fx[0][1] - Fxy[0][1] - Fy[0][1] + Fy[1][1];
	  c(i, j)[3][0] = -f(i,j) + f(1 + i,j) - Fx[1][0];
	  c(i, j)[3][1] = -f(i,1 + j) + f(1 + i,1 + j) - Fx[1][1];
	  c(i, j)[3][2] = -f(i,j) + f(i,1 + j) + f(1 + i,j) - f(1 + i,1 + j) - Fx[1][0] + Fx[1][1] - Fxy[1][0] - Fy[0][0] + Fy[1][0];
	  c(i, j)[3][3] = f(i,j) - f(i,1 + j) - f(1 + i,j) + f(1 + i,1 + j) + Fx[1][0] - Fx[1][1] + Fxy[1][1] + Fy[0][1] - Fy[1][1];
	}
      }
      assert(interpolates(f));
    }

    template<typename T>
    inline double interpolation_grid<T>::x_min() const{
      assert(std::is_sorted(begin(x), end(x)));
      return x.front();
    }

    template<typename T>
    inline double interpolation_grid<T>::x_max() const{
      assert(std::is_sorted(begin(x), end(x)));
      return x.back();
    }

    template<typename T>
    inline double interpolation_grid<T>::y_min() const{
      assert(std::is_sorted(begin(y), end(y)));
      return y.front();
    }

    template<typename T>
    inline double interpolation_grid<T>::y_max() const{
      assert(std::is_sorted(begin(y), end(y)));
      return y.back();
    }

    template<typename T>
    inline bool interpolation_grid<T>::empty() const{
      return x.empty();
    }

    template<typename T>
    inline T interpolation_grid<T>::operator()(double x_, double y_) const{
      const size_t x_idx = get_idx(x, x_);
      const size_t y_idx = get_idx(y, y_);
      return interpolate(x_, y_, x_idx, y_idx);
    }

    template<typename T>
    inline T interpolation_grid<T>::interpolate_bilinear(double x_, double y_) const{
      const size_t x_idx = get_idx(x, x_);
      const size_t y_idx = get_idx(y, y_);
      return interpolate_bilinear(x_, y_, x_idx, y_idx);
    }

    template<typename T>
    inline T interpolation_grid<T>::interpolate(
        double x_, double y_,
        size_t x_idx, size_t y_idx
    ) const{
      /*
       * Use bicubic interpolation.
       * The interpolating polynomial p = p_ij*x^i*y^j
       * is written in terms of rescaled variables tx, ty,
       * so that the corners of the grid cell
       * correspond to the four combinations tx, ty in {0, 1}
      */
      assert(x_ <= x_max());
      assert(y_ <= y_max());
      assert(x_idx < c.rows());
      assert(y_idx < c.columns());
      const double tx = (x_-x[x_idx])/(x[x_idx+1]-x[x_idx]);
      const double ty = (y_-y[y_idx])/(y[y_idx+1]-y[y_idx]);
      const double txb = 1 - tx;
      const double tyb = 1 - ty;
      auto const & C = c(x_idx, y_idx);
      return
	txb*(
	    tyb*C[0][0] + ty*(
		C[0][1] + tyb*(
		    tyb*C[0][2] + ty*C[0][3]
		)
	    )
	)
	+ tx*(
	    tyb*(C[1][0] + txb*(txb*C[2][0] + tx*C[3][0]))
	    + ty*(
		C[1][1] + tyb*(tyb*C[1][2] + ty*C[1][3])
		+ txb*(
		    txb*C[2][1] + tx*C[3][1]
		    + tyb*(
			txb*(tyb*C[2][2] + ty*C[2][3])
			+ tx*(tyb*C[3][2] + ty*C[3][3])
		    )
		)
	    )
	);
    }

    template<typename T>
    inline T interpolation_grid<T>::interpolate_bilinear(
        double x_, double y_,
        size_t x_idx, size_t y_idx
    ) const{
      /*
        use bilinear interpolation
        f(x, y) = f(x0, y0) + c1*delta_x + c2*delta_y
        + c3*delta_x*delta_y
      */
      return 1/((x[x_idx+1] - x[x_idx])*(y[y_idx+1] - y[y_idx]))*(
          + (x[x_idx+1] - x_)*(y[y_idx+1] - y_) * c(x_idx, y_idx)[0][0]
          + (x[x_idx] - x_)*(y_ - y[y_idx+1]) * c(x_idx+1, y_idx)[0][0]
          + (x_ - x[x_idx+1])*(y[y_idx] - y_) * c(x_idx, y_idx+1)[0][0]
          + (x_ - x[x_idx])*(y_ - y[y_idx]) * c(x_idx+1,y_idx+1)[0][0]
      );
    }

    template<typename T>
    inline size_t interpolation_grid<T>::get_idx(
        std::vector<double> const & grid_coord,
        double coord
    ) const{
      assert(std::is_sorted(begin(grid_coord), end(grid_coord)));
      auto it = std::lower_bound(
          grid_coord.begin(),
          grid_coord.end(),
          coord,
          std::less_equal<double>{}
      );
      return std::distance(begin(grid_coord), it) - 1;
    }

    template<typename T, typename U>
    inline bool is_roughly_same(double ep, T t, U u){
      using std::abs;
      if(t == T{}) return (abs)(u) < ep;
      return (abs)(t - u) < ep*(abs)(u);
    }

    template<typename T, typename U, typename V, typename... Rest>
    inline bool is_roughly_same(double ep, T t, U u, V v, Rest... r){
      return is_roughly_same(ep, t, u) && is_roughly_same(ep, u, v, r...);
    }

    template<typename T>
    inline bool interpolation_grid<T>::interpolates(
	matrix<T> const & f
    ) const{
#ifndef NDEBUG
      static constexpr double ep = 1e-14;
#endif
      for(size_t i = 0; i < f.rows()-1; ++i){
	for(size_t j = 0; j < f.columns()-1; ++j){
	  for(size_t di = 0; di < 2; ++di){
	    for(size_t dj = 0; dj < 2; ++dj){
	      assert(
		  is_roughly_same(
		      ep,
		      f(i+di, j+dj),
		      interpolate(x[i+di], y[j+dj], i, j)
		  )
	      );
	    }
	  }
	}
      }
      return true;
    }

  }
}
