/**
 * @file
 * @brief Definitions of mass schemes
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.4
 *
 * @section LICENSE
 *
 * Copyright 2014, 2016 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#pragma once
#include <array>
#include <string>
#include <ostream>

//! -
namespace QQbar_threshold{

  //! Names of mass schemes
  /**
   * If a shifted scheme is used in the evaluation of an expression
   * the mass is first converted to the pole scheme at the given order
   * and then inserted into the expression for the pole mass.
   *
   * For insertion schemes the leading-order conversion to the pole scheme
   * is used. Higher-order corrections in the relation to the pole mass
   * are treated as insertions of counterterms \cite Beneke:2014.
   *
   * @see is_shifted_mass
   */
  enum scheme_name{
    //! The potential-subtracted insertion scheme
    PS,
    //! The potential-subtracted shift scheme
    PSshift,
    //! The 1S insertion scheme
    OneS,
    //! The 1S shift scheme
    OneSshift,
    //! The pole scheme
    Pole,
    //! The \f$\overline{\rm MS}\f$ scheme
    MSshift,
    MSbarshift = MSshift,
    MSbar = MSshift,
    last_mass_scheme = MSbar
  };

  //! Mass scheme
  /**
   * A mass scheme is in general defined by its name and an associated
   * scale. Some schemes (Pole, 1S) do not have an explicit scale, so
   * the value of the corresponding member will be ignored
   */
  struct scheme{
    //! Name of the scheme
    scheme_name name;
    //! Associated scale
    double scale;
  };

  inline
  std::string to_string(scheme s){
    using std::to_string;
    static constexpr
      std::array<char const *, last_mass_scheme + 1> name_strings = {{
	"PS", "PSshift", "OneS", "OneSshift", "Pole", "MSbar"
      }};
    return
      std::string{'{'} + name_strings[s.name] + ", " + to_string(s.scale) + '}';
  }

  inline
  std::ostream& operator<<(std::ostream& os, scheme s){
    os << to_string(s);
    return os;
  }

  inline bool is_threshold_mass(scheme_name s){
    switch(s){
    case Pole:
    case PSshift:
    case OneSshift:
    case MSbar:
      return false;
    default: return true;
    }
  }

  inline bool is_threshold_mass(scheme const & s){
    return is_threshold_mass(s.name);
  }

  //! Determines whether the given scheme is a shifted scheme
  /**
   * @param s The name of the mass scheme
   * @returns \c true for shifted mass schemes, \c false otherwise
   */
  inline bool is_shifted_mass(scheme_name  s){
    switch(s){
    case PSshift:
    case OneSshift:
    case MSbar:
      return true;
    default: return false;
    }
  }

  //! Returns the corresponding shifted scheme
  /**
   * @param s The name of the original mass scheme
   * @returns The corresponding shifted scheme
   *
   * For example, this returns PSshift if the input scheme is PS.
   * If the input scheme is a shifted scheme or the pole scheme
   * the input scheme is returned.
   */
  inline scheme_name to_shifted_mass(scheme_name  s){
    switch(s){
    case PS: return PSshift;
    case OneS: return OneSshift;
    default: return s;
    }
  }

  //! Determines whether the given scheme is a shifted scheme
  /**
   * @param s The mass scheme
   * @returns \c true for shifted mass schemes, \c false otherwise
   */
  inline bool is_shifted_mass(scheme const & s){
    return is_shifted_mass(s.name);
  }
}
