/**
 * @file
 * @brief Options for \f$q\bar{q}\f$ functions
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.4
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015, 2016 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */

#pragma once
#include <cassert>
#include <array>
#include <limits>
#include <complex>
#include <type_traits>
#include <ostream>
#include <string>

#include "constants.hpp"
#include "mass_schemes.hpp"
#include "alpha_s.hpp"

namespace QQbar_threshold{
  typedef std::complex<double> complex;

  namespace pert_order{
    enum order{
      LO = 0,
      NLO,
      N2LO,
      NNLO = N2LO,
      N3LO,
      NNNLO = N3LO
    };

    inline
    order operator++(order & o){
      return o = static_cast<order>(o + 1);
    }

    inline
    order operator--(order & o){
      return o = static_cast<order>(o - 1);
    }

    std::string to_string(order o);

    std::ostream& operator<<(std::ostream& os, order o);

  }
  using namespace pert_order;

  enum class SM;

  inline
  constexpr SM operator|(SM a, SM b){
    typedef std::underlying_type<SM>::type SM_type;
    return static_cast<SM>(static_cast<SM_type>(a) | static_cast<SM_type>(b));
  }

  inline
  SM & operator|=(SM & a, SM b){
    return a = a | b;
  }

  inline
  constexpr SM operator&(SM a, SM b){
    typedef std::underlying_type<SM>::type SM_type;
    return static_cast<SM>(static_cast<SM_type>(a) & static_cast<SM_type>(b));
  }

  //! Standard Model effects beyond QCD.
  /**
   * Effects can be combined with the | operator
   * e.g. @code
   * options opt;
   * opt.beyond_QCD = SM::QED | SM::Higgs;
   * @endcode
   * activates both QED and Higgs effects.
   */
  enum class SM{
    //! No corrections beyond QCD
    none = 0,
    //! Corrections due to photons
    QED = 1 << 0,
    //! Corrections due to the physical Higgs boson
    Higgs = 1 << 1,
    //! \internal auxiliary value
    QEDHiggs = QED|Higgs,
    //! All Standard Model corrections
    all = (1 << 3) - 1
  };

  inline
  constexpr bool operator!(SM a){
    return a == SM::none;
  }

  std::string to_string(SM s);

  std::ostream& operator<<(std::ostream& os, SM s);

  //! Production channels
  enum class production_channel{
    //! Production only via a virtual photon
    photon_only = 0,
    //! Production only via S-wave photon or Z
      S_wave_only,
    //! All possible production channels
      all
  };

  std::string to_string(production_channel p);

  std::ostream& operator<<(std::ostream& os, production_channel p);


  //! Multiplicative coefficients for the potentials and current matching coefficients.
  /**
   * E.g. setting
   * @code
   * options opt;
   * opt.contributions.v_delta = {{0.0, 1.0}};
   * @endcode
   * implies that corrections due to the leading-order delta potential
   * are multiplied by zero, i.e. effectively discarded,
   * but corrections from the next-to-leading delta potential are kept.
   *
   * The contributions are defined as in \cite Beneke:2013jia.
   */
  struct contribution_factors{
    //! Corrections \f$\hat{\cal V}_C^{(1)},\hat{\cal V}_C^{(2)},\hat{\cal V}_C^{(3)}\f$ to the colour Coulomb potential
    std::array<double,3> v_Coulomb = {{1.,1.,1.}};
    //! Corrections \f$\hat{\cal V}_{1/m^2}^{(1)},\hat{\cal V}_{1/m^2}^{(2)}\f$ to the \f$\delta\f$ (or \f$1/m^2\f$) potential
    std::array<double,2> v_delta = {{1.,1.}};
//! Corrections \f$\hat{\cal V}_{1/m}^{(1)},\hat{\cal V}_{1/m}^{(2)}\f$ to the \f$1/r^2\f$ (or \f$1/m\f$) potential
    std::array<double,2> v_r2inv = {{1.,1.}};
    //! Corrections \f$\hat{\cal V}_{p}^{(1)},\hat{\cal V}_{p}^{(2)}\f$ to the momentum dependent potential
    std::array<double,2> v_p2 = {{1.,1.}};
    //! Kinetic corrections
    std::array<double,1> v_kinetic = {{1.}};
    //! Ultrasoft corrections
    std::array<double,1> ultrasoft = {{1.}};
    //! Higgs potential
    std::array<double,1> v_Higgs = {{1.}};
    //! Corrections due to the classical (QED) Coulomb potential
    std::array<double,1> v_QED_Coulomb = {{1.}};
    //! Corrections to the hard matching coefficient \f$c_v\f$
    std::array<double,3> cv = {{1.,1.,1.}};
    //! Higgs corrections to the hard matching coefficient \f$c_v\f$
    std::array<double,2> cv_Higgs = {{1.,1.}};
    //! QED corrections to hard matching with vector coupling to leptons
    std::array<double,1> Cv_QED = {{1.}};
    //! W,Z corrections to hard matching with vector coupling to leptons
    std::array<double,1> Cv_WZ = {{1.}};
    //! QED corrections to hard matching with axialvector coupling to leptons
    std::array<double,1> Ca_QED = {{1.}};
    //! W,Z corrections to hard matching with axialvector coupling to leptons
    std::array<double,1> Ca_WZ = {{1.}};
    //! Sub-leading hard matching coefficient \f$d_v\f$
    std::array<double,2> dv = {{1.,1.}};
    //! P-wave hard matching coefficient \f$c_a\f$
    std::array<double,1> ca = {{1.}};
  };

  struct alpha_s_options{
    double mu0;
    double alpha_s;
  };

  std::string to_string(alpha_s_options o);

  static_assert(
      std::numeric_limits<double>::has_quiet_NaN,
      "Quiet NaN not available"
  );

  //! Options for \f$q\bar{q}\f$ functions
  /**
   * The default settings depend on the specific function.
   * Members that are irrelevant to the considered function are ignored.
   */
  struct options{
    /**
     *  Specifies multiplicative factors for various contributions
     *
     * Option for cross section, energy level and residue functions
     * @see contribution_factors
     */
    contribution_factors contributions;
    /**
     * Specifies value of the strong coupling at the scale \ref mZ
     *
     * Option for cross section, energy level and residue functions.
     * @see mZ, alpha_s_mu0
     */
    double alpha_s_mZ = QQbar_threshold::alpha_s_mZ;
    /**
     * alpha_s_mu0 = {mu0, alpha_s} specifies a reference scale mu0
     * and the value alpha_s of the strong coupling constant at that scale.
     * If set, this option overrides the \ref alpha_s_mZ option.
     *
     * For bottom-related functions decoupling to the four-flavour theory
     * will be performed if mu0 is above the decoupling scale \ref mu_thr.
     * For lower values of mu0 it is assumed that alpha_s is given
     * in the four-flavour theory.
     *
     * Option for cross section, energy level and residue functions.
     * @see mZ, alpha_s_mZ
     */
    alpha_s_options alpha_s_mu0 = {
      std::numeric_limits<double>::quiet_NaN(),
      std::numeric_limits<double>::quiet_NaN()
    };
    /**
     * Specifies the value of the Higgs boson mass.
     *
     * Option for cross section, energy level and residue functions.
     *
     */
    double m_Higgs = QQbar_threshold::m_Higgs;
    /**
     * Specifies  a multiplier for the quark Yukawa coupling.
     *
     * Option for cross section, energy level and residue functions.
     *
     */
    double Yukawa_factor = 1;
    /**
     * Specifies an invariant mass cut for the non-resonant contribution
     *
     * TODO: check
     * The invariant mass of the final-state particles will be restricted
     * to the region between m - invariant_mass_cut and m + invariant_mass_cut.
     * By default, the loosest possible cut invariant_mass_cut = m - mW is taken,
     * which yields the total cross section.
     *
     * Option for cross section functions.
     *
     */
    double invariant_mass_cut = std::numeric_limits<double>::quiet_NaN();
    /**
     * Specifies the value of the light (e.g. charm or bottom) quark mass
     *
     * Option for all bottom- and top-related function.
     * For the cross section functions, the implementation only includes
     * corrections up to order = 1. For other functions also the corrections
     * for order = 2 are included.
     * @see double_light_insertion
     */
    double ml = 0;
    /**
     * Third-order coefficient for conversion between pole and MS scheme
     *
     * Specifies the non-logarithmic coefficient in the conversion between
     * the pole and the MS scheme at order == N3LO.
     *
     * r4 is mostly equivalent to \f$c_m^{(4)}(\mu = m(m))\f$ in
     * \cite Marquard:2015qpa, but r4 refers to the theory with nl
     * as opposed to nl + 1 active flavours.
     *
     * Option for cross section, energy level and residue functions.
     */
    double r4 = std::numeric_limits<double>::quiet_NaN();
    /**
     * Specifies the value of the running QED coupling constant.
     *
     * Option for cross section, energy level and residue functions
     */
    double alpha;
    /**
     * Specifies the scale for the running QED coupling constant.
     *
     * Option for cross section, energy level and residue functions
     */
    double mu_alpha;
    /**
     * Specifies the number of bound states that are resummed into the Green function.
     *
     * Option for cross section functions.
     *
     * @see alpha_s, \cite Beneke:2014, \cite Beneke:2014pta
     */
    int resum_poles = 6;
    /**
     * Specifies the corrections beyond QCD that should be taken into account.
     *
     * Option for cross section, energy level and residue functions.
     * @see SM
     */
    SM beyond_QCD;
    /**
     * Specifies the renormalisation scheme and scale of the quark mass
     *
     * Option for all bottom- or top-related functions
     * @see mass_scheme
     */
    scheme mass_scheme;
    /**
     * Specifies which production channels are taken into account
     *
     * Option for cross section functions.
     * @see production_channel
     */
    production_channel production = production_channel::all;
    /**
     * Specifies whether only the resonant production should be computed
     *
     * Option for cross section functions.
     */
    bool resonant_only;
    /**
     * Specifies treatment of the global factor 1/s in the polarisation function
     *
     * If set to \c true, 1/s is expanded around the threshold center-of-mass
     * energy, i.e. the relation between the polarisation function and the
     * PNRQCD Green function at leading order reads
     * \f$\Pi = \frac{N_c}{2m^2} G(E)\f$. If set to false, the factor
     * is kept in unexpanded form, so that
     * \f$\Pi = \frac{2 N_c}{s} G(E)\f$ ar leading order.
     *
     * Option for cross section and residue functions.
     */
    bool expand_s;
    /**
     * Specifies whether effects from a non-zero light quark mass on double potential insertions are taken into account.
     *
     * Setting this option to \c true usually has a small numerical impact on the
     * final result and leads to a considerable slowdown. The current
     * implementation is comparably fast, but not very accurate. Furthermore,
     * it is **not thread safe**.
     *
     * Option for energy level and residue functions.
     */
    bool double_light_insertion = false;
  };

  std::string to_string(options const & opt);

  struct scales{
    double soft;
    double weak;
  };

  struct top_properties{
    double m;
    double width;
  };

  //! Describe internal settings for bottom-related functions
  /**
   * @param mu    renormalisation scale
   * @param m     quark mass
   * @param o     perturbative order (LO, NLO, N2LO, or N3LO)
   * @param opt   further options
   * @returns A string containing the names and values of internal variables
   *
   * This function is intended for debugging purposes.
   */
  std::string bottom_internal_settings(
      double mu, double m, order o, options opt
  );
  //! Describe internal settings for bottom-related functions
  /**
   * @param sqrt_s center-of-mass energy
   * @param mu     renormalisation scale
   * @param m      bottom quark mass
   * @param o      perturbative order (LO, NLO, N2LO, or N3LO)
   * @param opt    further options
   * @returns A string containing the names and values of internal variables
   *
   * This function is intended for debugging purposes.
   */
  std::string bottom_internal_settings(
      double sqrt_s, double mu, double m, order o, options opt
  );
  //! Describe internal settings for top-related functions
  /**
   * @param mu    renormalisation scale
   * @param m     quark mass
   * @param o     perturbative order (LO, NLO, N2LO, or N3LO)
   * @param opt   further options
   * @returns A string containing the names and values of internal variables
   *
   * This function is intended for debugging purposes.
   */
  std::string top_internal_settings(
      double mu, double m, order o, options opt
  );
  //! Describe internal settings for top-related functions
  /**
   * @param sqrt_s center-of-mass energy
   * @param mu     renormalisation scales
   * @param M      top quark mass and width
   * @param o      perturbative order (LO, NLO, N2LO, or N3LO)
   * @param opt    further options
   * @returns A string containing the names and values of internal variables
   *
   * This function is intended for debugging purposes.
   */
  std::string top_internal_settings(
      double sqrt_s, scales mu, top_properties M, order o, options opt
  );
  std::string top_internal_settings(
      double sqrt_s, double mu, top_properties M, order o, options opt
  );

  std::ostream& operator<<(std::ostream& os, options const & obt);

  //! Default option settings for bottom-related functions
  /**
   * The default settings are:
   *
   * - options::contributions = (all set to 1)
   * - options::alpha_s_mZ = #alpha_s_mZ
   * - options::m_Higgs = \ref m_Higgs
   * - options::Yukawa_factor = 0
   * - options::invariant_mass_cut = 0
   * - options::ml = 0
   * - options::r4 = 1220.3 (for four active flavours)
   * - options::alpha = \ref alpha_Y;
   * - options::mu_alpha = \ref mu_alpha_Y;
   * - options::resum_poles = 6;
   * - options::beyond_QCD = SM::QED;
   * - options::mass_scheme = {\ref PSshift, mu_f_bottom};
   * - options::production = production_channel::all;
   * - options::expand_s = false;
   * - options::double_light_insertion = false;
   */
  options bottom_options();
  //! Default option settings for top-related functions
  /**
   * The default settings are:
   *
   * - options::contributions = (all set to 1)
   * - options::alpha_s_mZ = #alpha_s_mZ
   * - options::m_Higgs = \ref m_Higgs
   * - options::Yukawa_factor = 1
   * - options::invariant_mass_cut = m -\ref mW
   * - options::ml = 0
   * - options::r4 = 824.12 (for five active flavours)
   * - options::alpha = \ref alpha_mZ;
   * - options::mu_alpha = \ref mZ;
   * - options::resum_poles = 6;
   * - options::beyond_QCD = SM::all;
   * - options::mass_scheme = {\ref PSshift, mu_f_bottom};
   * - options::production = production_channel::all;
   * - options::expand_s = true;
   * - options::double_light_insertion = false;
   */
  options top_options();

  /** \internal
   *  Internal struct to pass around additional parameters
   */
  struct parameters{
      contribution_factors contributions;
      complex E_shifted;
      double s;
      double m;
      double width;
      double mu;
      double alpha_s;
      double m_Higgs;
      double Yukawa_t2;
      double invariant_mass_cut;
      double ml;
      double e_q;
      //! \internal Z couplings a_q, v_q to quark
      std::array<double, detail::last_coupling + 1> c_q;
      double r4;
      double mu_w;
      double alpha;
      double mu_alpha;
      int order;
      int nl;
      int resum_poles;
      SM beyond_QCD;
      scheme mass_scheme;
      production_channel production;
      bool resonant_only;
      bool expand_s;
      bool double_light_insertion;
  };

  parameters gen_bottom_parameters(
      double mu, double m, order o,
      options opt
  );
  parameters gen_bottom_parameters(
      double sqrt_s, double mu, double m, order o,
      options opt
  );
  parameters gen_bottom_parameters(
      double sqrt_s, scales mu, double m, order o,
      options opt
  );
  parameters gen_top_parameters(
      double mu, double m, order o,
      options opt
  );
  parameters gen_top_parameters(
      double sqrt_s, double mu, top_properties M, order o,
      options opt
  );
  parameters gen_top_parameters(
      double sqrt_s, scales mu, top_properties M, order o,
      options opt
  );

  std::string to_string(parameters const & p);

}
