/**
 * @file
 * @brief Implementation of pole resummation
 *
 * @see pole_resummation.hpp
 *
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.4
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015, 2016 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#include "pole_resummation.hpp"

#include <vector>

#include "energy_levels.hpp"
#include "residues.hpp"

namespace QQbar_threshold{

  namespace detail{
    complex resum_pole_Pi(int n, parameters const & p){
      if(p.order == 0) return 0;
      const double E_N = QQbar_energy_level_no_shift(n, p);
      complex resummed_pole =
	QQbar_residue(n, p)/(E_N - p.E_shifted);
      complex delta_E = E_LO(n, p) - p.E_shifted;
      complex E_ratio = E_LO(n, p)/delta_E;
      std::vector<double> z(p.order+1);
      std::vector<double> e(p.order+1);
      switch(p.order){
      case 3:
	z[3] = z3(n, p);
	e[3] = e3(n, p);
      case 2:
	z[2] = z2(n, p);
	e[2] = e2(n, p);
      case 1:
	z[1] = z1(n, p);
	e[1] = e1(n, p);
      default:;
      }
      complex expanded_pole = 1;
      switch(p.order){
      case 3: expanded_pole +=
	  + z[3]
	  + E_ratio*(
	      - e[3]
	      - e[2]*z[1]
	      - e[1]*z[2]
	      + E_ratio*(
		  + e[1]*(
		      + 2*e[2]
		      + e[1]*(z[1] - E_ratio*e[1])
		  )
	      )
	  );
      case 2: expanded_pole +=
	  + z[2]
	  + E_ratio*(
	      - e[2]
	      + e[1] * (- z[1] + E_ratio*e[1])
	  );
      case 1: expanded_pole += z[1] - E_ratio*e[1];
      default:;
      }
      expanded_pole *= Z_LO(n, p)/delta_E;
      if(!p.expand_s){
        const double E_N_RS = is_threshold_mass(p.mass_scheme)?
          QQbar_energy_level(n, p, p.m):E_N;
        expanded_pole *= pow(1 + E_N_RS/(2*p.m), -2);
      }
      return resummed_pole - expanded_pole;
    }

    template<>
    complex resum_pole_GF<0>(int n, parameters const & p){
      const double E_N = QQbar_energy_level_no_shift(n, p);
      return Z_LO(n, p)*(1./(E_N - p.E_shifted) - 1./(E_LO(n, p) - p.E_shifted));
    }

    template<>
    complex resum_pole_GF<1>(int n, parameters const & p){
      const double E_N = QQbar_energy_level_no_shift(n, p);
      const double f1_N = f1(n, p);
      const complex delta_E = E_LO(n, p) - p.E_shifted;
      const complex E_ratio = E_LO(n, p)/delta_E;
      const complex resummed_pole = f1_N/(E_N - p.E_shifted);
      const complex expanded_pole = (f1_N - e1(n, p)*E_ratio)/delta_E;
      return Z_LO(n, p)*(resummed_pole - expanded_pole);
    }
  }
}
