/**
 * @file
 * @brief Top quark decay width
 *
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.4
 *
 * @section LICENSE
 *
 * Copyright 2015, 2016 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */

#include <stdexcept>

#include "parameters.hpp"
#include "scheme_conversion.hpp"
#include "zeta.hpp"
#include "dilog.hpp"
#include "constants.hpp"
#include "QCD/colour.hpp"
#include "QCD/beta.hpp"

//! -
namespace QQbar_threshold{
  /**
   * Expressions for Born and 1 loop taken from \cite Fischer:2000kx
   * (original ref: \cite Jezabek:1993wk)
   * 2 loop correction from \cite Blokland:2005vq
   *
   * V_tb = 1
   */
  namespace{
    using zeta::pi;
    using zeta::zeta_2;
    using zeta::zeta_3;
    using zeta::zeta_4;
    using dilog::Li2;

    double lambda(double x2, double y2){
      return x2*x2 + (y2 - 1)*(y2 - 1) - 2*x2*(1 + y2);
    }

    //! Global prefactor
    /**
     * eq. (59) in \cite Blokland:2005vq
     * |V_tb| = 1
     */
    double Gamma_0(double mt){
      return G_F*mt*mt*mt/(8*sqrt(2)*pi);
    }

    //! Born contribution
    /**
     * cf. eq. (5) in \cite Fischer:2000kx
     * note different normalisation (eq. (58) in \cite Blokland:2005vq)
     */
    double X_0(double x, double y){
      const double x2 = x*x;
      const double y2 = y*y;
      return sqrt(lambda(x2, y2))*((1 - y2)*(1 - y2) + x2*(1 - 2*x2 + y2));
    }

    //! 1 loop contribution
    /**
     * cf. eq. (14) in \cite Fischer:2000kx
     * note different normalisation (eq. (58) in \cite Blokland:2005vq)
     * X_1 = X_0*(\hat{\Gamma} - 1)*pi/alpha_s
     */
    double X_1(double x){
      const double x2 = x*x;
      return
        X_0(x, 0)*QCD::c_F/2*(
            - 8*(Li2(x) + Li2(-x)) - 4*log(x)*log(1 - x2) - 4*zeta_2
            + (
                - (5 + 4*x2)*log(1 - x2)
                + (
                    + 5./2. + x2*(
                        + 9./2. - 3*x2 - 4*(1 + x2)*(1 - 2*x2)*log(x)/(1 - x2)
                    )
                )/(1 - x2)
            )/(1 + 2*x2)
        );
    }

    //! 2 loop light flavour contribution
    double X_L(double x2, double log_x2){
      return
        - 4./9. + 23./18.*zeta_2 + zeta_3
        + x2*(
            - 19./6. + 4./3.*zeta_2
            + x2*(
                745./72. - 31./6.*zeta_2 - 3*zeta_3 - 7./4.*log_x2
                + x2*(
                    - 5839./648. + 14./9.*zeta_2 + 2*zeta_3 + 5./3.*log_x2
                    + x2*(
                        4253./8640. + 3./2.*zeta_2 - 17./144.*log_x2
                        + x2*(
                            - 689./27000. + 2./5.*zeta_2 - 7./900.*log_x2
                            + x2*(
                                - 13187./181440. + zeta_2/6. + 1./48.*log_x2
                                + x2*(-2282381./37044000. + 3./35.*zeta_2 + 2263./88200.*log_x2)
                            )
                        )
                    )
                )
            )
        );
    }

    //! 2 loop heavy flavour contribution
    double X_H(double x2){
      return 12991./1296. - (53*zeta_2)/9 - zeta_3/3 + x2*(-35./108. - (8*zeta_2)/3 + 4*zeta_3 + x2*(-6377./432. + (25*zeta_2)/3 + x2*(319./27. + x2*(76873./8640. + x2*(237107./27000. - (16*zeta_2)/3) - (16*zeta_2)/3) - (62*zeta_2)/9 - (2*zeta_3)/3) + zeta_3));
    }

    //! 2 loop abelian contribution
    double X_A(double x2, double log_x2){
      return 5 - (53*zeta_3)/8 - (11*zeta_4)/8 + x2*(-73./8. + (123*zeta_2)/4 - 41*zeta_4 + x2*(-7537./288. + (115*log_x2)/48 + (295*zeta_3)/32 - (191*zeta_4)/8 + x2*(16499./864. - (367*log_x2)/144 - (7*zeta_3)/2 + x2*(-1586479./259200. + log_x2*(31979./17280. - (3*zeta_2)/8) + (2951*zeta_2)/1152 + (9*zeta_3)/2 + x2*(-11808733./6480000. + log_x2*(13589./27000. - zeta_2/10) + (37*zeta_2)/400 + (6*zeta_3)/5)) + (21*zeta_4)/4 + zeta_2*(-407./36. + (10*log_x2)/3 - 6*log(2))) + zeta_2*(523./16. - (15*log_x2)/8 - (81*log(2))/8))) + zeta_2*(-119./8. + (57*log(2))/2);
    }

    //! 2 loop non-abelian contribution
    double X_NA(double x2, double log_x2){
      return 521./576. + (9*zeta_3)/16 + (11*zeta_4)/16 + zeta_2*(505./144. - (57*log(2))/4) + x2*(91./48. + (329*zeta_2)/24 - (39*zeta_4)/2 + x2*(-12169./576. + (73*log_x2)/16 + (377*zeta_3)/64 - (385*zeta_4)/16 + zeta_2*(2171./96. - (9*log_x2)/16 + (81*log(2))/16) + x2*(13685./864. - (1121*log_x2)/432 + x2*(-420749./103680. + log_x2*(11941./6912. - (9*zeta_2)/16) - (3263*zeta_2)/2304 + x2*(-4868261./12960000. + log_x2*(153397./216000. - (3*zeta_2)/20) - (557*zeta_2)/800 - (3*zeta_3)/10) - (9*zeta_3)/8) - (19*zeta_3)/4 + (43*zeta_4)/8 + zeta_2*(-47./12. - log_x2 + 3*log(2)))));
    }

    //! Complete 2 loop contribution
    /**
     * cf. eq. (60) in \cite Blokland:2005vq
     * nh = 1
     */
    double X_2(double x, int nl){
      const double x2 = x*x;
      const double log_x2 = log(x2);
      return QCD::c_F*(
          + QCD::T_f*(nl*X_L(x2, log_x2) + X_H(x2))
          + QCD::c_F*X_A(x2, log_x2)
          + QCD::c_A*X_NA(x2, log_x2)
      );
    }

    //! Threshold mass correction
    /*
     * In contrast to other threshold mass corrections
     * the factor delta_m/m is not included here
     */
    double X_0_TM(double x){
      const double x2 = x*x;
      return 12*x2*x2*(1 - x2);
    }

    double top_width(order o, parameters const & p){
      const double x = mW/p.m;
      const double y = p.ml/p.m;
      double result = 0;
      switch(o){
      case 2:
        result += pow(p.alpha_s/pi, 2)*(
          + X_2(x, p.nl)
          + 2*log(p.mu/p.m)*QCD::beta(0, p.nl)*X_1(x)
        );
        // Approximate EW correction (cf. \cite Denner:1990ns,Eilam:1991iz)
        result += 0.017*X_0(x, y);
        if(is_threshold_mass(p.mass_scheme)){
          result += detail::delta_m(0, p)/p.m*(
              3*X_0(x, 0) + X_0_TM(x)
          );
        }
      case 1:
        result += p.alpha_s/pi*X_1(x);
      case 0:
        result += X_0(x, y);
      break;
      default:
      throw std::out_of_range(
	  "order " + std::to_string(p.order)
	  + " out of range (0,...,2) for top width"
      );
      }
      return Gamma_0(p.m)*result;
    }
  }

  double top_width(double mu, double m, order o, options opt){
    return top_width(o, gen_top_parameters(mu, m, N3LO, opt));
  }
}
